/** @file
 *
 * Module to dynamically load libvdeplug and load all symbols
 * which are needed by VirtualBox - header file.
 */

/*
 * Copyright (C) 2008-2015 Oracle Corporation
 *
 * This file is part of VirtualBox Open Source Edition (OSE), as
 * available from http://www.virtualbox.org. This file is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License (GPL) as published by the Free Software
 * Foundation, in version 2 as it comes in the "COPYING" file of the
 * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
 *
 * The contents of this file may alternatively be used under the terms
 * of the Common Development and Distribution License Version 1.0
 * (CDDL) only, as it comes in the "COPYING.CDDL" file of the
 * VirtualBox OSE distribution, in which case the provisions of the
 * CDDL are applicable instead of those of the GPL.
 *
 * You may elect to license modified versions of this file under the
 * terms and conditions of either the GPL or the CDDL or both.
 */

#ifndef ___VBox_VDEPlug_h
#define ___VBox_VDEPlug_h

#define LIBVDEPLUG_INTERFACE_VERSION 1

#define vde_open(vde_switch, descr, open_args) \
    vde_open_real((vde_switch), (descr), LIBVDEPLUG_INTERFACE_VERSION, (open_args))

/* Declarations of the functions that we need from the library */
#define VDEPLUG_GENERATE_HEADER

#include <VBox/VDEPlugSymDefs.h>

#undef VDEPLUG_GENERATE_HEADER

#endif /* ___VBox_VDEPlug_h not defined */
/* vi: set tabstop=4 shiftwidth=4 expandtab: */
